# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Support utilities for ds105 example, mostly path management.

The purpose of separating these is to keep the main example code as readable as
possible and focused on the experimental modeling and analysis, rather than on
local file management issues.

Requires matplotlib
"""

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from __future__ import print_function # Python 2/3 compatibility

# Stdlib
import os
from os import makedirs, listdir
from os.path import exists, abspath, isdir, join as pjoin, splitext

# Third party
import numpy as np
from matplotlib.mlab import csv2rec

# From NIPY
from nipy.io.api import load_image

#-----------------------------------------------------------------------------
# Globals
#-----------------------------------------------------------------------------

# We assume that there is a directory holding the data and it's local to this
# code.  Users can either keep a copy here or a symlink to the real location on
# disk of the data.
DATADIR = 'ds105_data'

# Sanity check
if not os.path.isdir(DATADIR):
    e="The data directory %s must exist and contain the ds105 data." % DATADIR
    raise IOError(e)

#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------

# Path management utilities
def load_image_ds105(*path):
    """Return a NIPY image from a set of path components.
    """
    return load_image(pjoin(DATADIR, *path))


def subj_des_con_dirs(design, contrast, subjects=range(1,7)):
    """Return a list of subject directories with this `design` and `contrast`

    Parameters
    ----------
    design : {'standard'} 
    contrast : str
    subjects : list, optional
        which subjects

    Returns
    -------
    con_dirs : list
        list of directories matching `design` and `contrast`
    """
    rootdir = DATADIR
    con_dirs = []
    for s in range(nsub):
        f = pjoin(rootdir, "sub%03d" % s, "model", design, "fixed", contrast)
        if isdir(f):
            con_dirs.append(f)
    return con_dirs


def path_info_run(subj, run, design='standard'):
    """Construct path information dict for current subject/run.

    Parameters
    ----------
    subj : int
        subject number (1..6 inclusive)
    run : int
        run number (1..12 inclusive).
    design : str, optional
        which design to use, defaults to 'standard'
    Returns
    -------
    path_dict : dict
        a dict with all the necessary path-related keys, including 'rootdir',
        and 'design', where 'design' can have values 'event' or 'block'
        depending on which type of run this was for subject no `subj` and run no
        `run`
    """
    path_dict = {'subj': subj, 'run': run, 'design':design}
    rootdir = pjoin(DATADIR, "sub%(subj)03d", "model", "%(design)s") % path_dict
    path_dict['rootdir'] = rootdir
    path_dict['fsldir'] = pjoin(DATADIR, "sub%(subj)03d", "model", "model001") % path_dict
    return path_dict


def path_info_design(subj, design):
    """Construct path information dict for subject and design.

    Parameters
    ----------
    subj : int
        subject number (1..6 inclusive)
    design : {'standard'}
        type of design

    Returns
    -------
    path_dict : dict
        having keys 'rootdir', 'subj', 'design'
    """
    path_dict = {'subj': subj, 'design': design}
    rootdir = pjoin(DATADIR, "sub%(subj)03d", "model", "%(design)s") % path_dict
    path_dict['rootdir'] = rootdir
    path_dict['fsldir'] = pjoin(DATADIR, "sub%(subj)03d", "model", "model001") % path_dict
    return path_dict


def results_table(path_dict):
    """ Return precalculated results images for subject info in `path_dict`

    Parameters
    ----------
    path_dict : dict
        containing key 'rootdir'

    Returns
    -------
    rtab : dict
        dict with keys given by run directories for this subject, values being a
        list with filenames of effect and sd images.
    """
    # Which runs correspond to this design type?
    rootdir = path_dict['rootdir']
    runs = filter(lambda f: isdir(pjoin(rootdir, f)),
                  ['results_run%03d' % i for i in range(1,13)] )

    # Find out which contrasts have t-statistics,
    # storing the filenames for reading below

    results = {}

    for rundir in runs:
        rundir = pjoin(rootdir, rundir)
        for condir in listdir(rundir):
            for stat in ['sd', 'effect']:
                fname_effect = abspath(pjoin(rundir, condir, 'effect.nii'))
                fname_sd = abspath(pjoin(rundir, condir, 'sd.nii'))
            if exists(fname_effect) and exists(fname_sd):
                results.setdefault(condir, []).append([fname_effect,
                                                       fname_sd])
    return results


def get_experiment(path_dict):
    """Get the record arrays for the experimental design.

    Parameters
    ----------
    path_dict : dict
        containing key 'rootdir', 'run', 'subj'

    Returns
    -------
    experiment, initial : Two record arrays.

    """
    # The following two lines read in the .csv files
    # and return recarrays, with fields
    # experiment: ['time', 'sentence', 'speaker']
    # initial: ['time', 'initial']

    rootdir = path_dict['rootdir']
    if not exists(pjoin(rootdir, "experiment_run%(run)03d.csv") % path_dict):
        e = "can't find design for subject=%(subj)d,run=%(subj)d" % path_dict
        raise IOError(e)

    experiment = csv2rec(pjoin(rootdir, "experiment_run%(run)03d.csv") % path_dict)

    return experiment


def get_fmri(path_dict):
    """Get the images for a given subject/run.

    Parameters
    ----------
    path_dict : dict
        containing key 'fsldir', 'run'

    Returns
    -------
    fmri : ndarray
    anat : NIPY image
    """
    fmri_im = load_image(
        pjoin("%(fsldir)s/task001_run%(run)03d.feat/filtered_func_data.nii.gz") % path_dict)
    return fmri_im


def ensure_dir(*path):
    """Ensure a directory exists, making it if necessary.

    Returns the full path."""
    dirpath = pjoin(*path)
    if not isdir(dirpath):
        makedirs(dirpath)
    return dirpath


def output_dir(path_dict, tcons, fcons):
    """Get (and make if necessary) directory to write output into.

    Parameters
    ----------
    path_dict : dict
        containing key 'rootdir', 'run'
    tcons : sequence of str
        t contrasts
    fcons : sequence of str
        F contrasts
    """
    rootdir = path_dict['rootdir']
    odir = pjoin(rootdir, "results_run%(run)03d" % path_dict)
    ensure_dir(odir)
    for n in tcons:
        ensure_dir(odir,n)
    for n in fcons:
        ensure_dir(odir,n)
    return odir

def compare_results(subj, run, other_root, mask_fname):
    """ Find and compare calculated results images from a previous run

    This scipt checks that another directory containing results of this same
    analysis are similar in the sense of numpy ``allclose`` within a brain mask.

    Parameters
    ----------
    subj : int
        subject number (1..6)
    run : int
        run number (1..12)
    other_root : str
        path to previous run estimation
    mask_fname:
        path to a mask image defining area in which to compare differences
    """
    # Get information for this subject and run
    path_dict = path_info_run(subj, run)
    # Get mask
    msk = load_image(mask_fname).get_data().copy().astype(bool)
    # Get results directories for this run
    rootdir = path_dict['rootdir']
    res_dir = pjoin(rootdir, 'results_run%03d' % run)
    if not isdir(res_dir):
        return
    for dirpath, dirnames, filenames in os.walk(res_dir):
        for fname in filenames:
            froot, ext = splitext(fname)
            if froot in ('effect', 'sd', 'F', 't'):
                this_fname = pjoin(dirpath, fname)
                other_fname = this_fname.replace(DATADIR, other_root)
                if not exists(other_fname):
                    print(this_fname, 'present but ', other_fname, 'missing')
                    continue
                this_arr = load_image(this_fname).get_data()
                other_arr = load_image(other_fname).get_data()
                ok = np.allclose(this_arr[msk], other_arr[msk])
                if not ok and froot in ('effect', 'sd', 't'): # Maybe a sign flip
                    ok = np.allclose(this_arr[msk], -other_arr[msk])
                if not ok:
                    print('Difference between', this_fname, other_fname)


def compare_all(other_root, mask_fname):
    """ Run results comparison for all subjects and runs """
    for subj in range(1,7):
        for run in range(1, 13):
            compare_results(subj, run, other_root, mask_fname)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Build helpers for setup.py

Includes package dependency checks, and code to build the documentation

To build the docs, run::

    python setup.py build_sphinx

This module has to work for python 2 and python 3.
"""
from __future__ import with_statement

# Standard library imports
import sys
import os
from os.path import join as pjoin, dirname
import zipfile
import warnings
import shutil
from distutils.cmd import Command
from distutils.command.clean import clean
from distutils.command.build_py import build_py
from distutils.version import LooseVersion
from distutils.dep_util import newer_group
from distutils.errors import DistutilsError

try:
    from ConfigParser import ConfigParser
except ImportError:  # Python 3
    from configparser import ConfigParser

from numpy.distutils.misc_util import appendpath
from numpy.distutils import log

# Sphinx import
try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    have_sphinx = False
else:
    have_sphinx = True

# Get project related strings.  Please do not change this line to use
# execfile because execfile is not available in Python 3
_info_fname = pjoin('nipy', 'info.py')
INFO_VARS = {}
exec(open(_info_fname, 'rt').read(), {}, INFO_VARS)

DOC_BUILD_DIR = os.path.join('build', 'html')

CYTHON_MIN_VERSION = INFO_VARS['CYTHON_MIN_VERSION']

################################################################################
# Distutils Command class for installing nipy to a temporary location. 
class TempInstall(Command):
    temp_install_dir = os.path.join('build', 'install')

    def run(self):
        """ build and install nipy in a temporary location. """
        install = self.distribution.get_command_obj('install')
        install.install_scripts = self.temp_install_dir
        install.install_base    = self.temp_install_dir
        install.install_platlib = self.temp_install_dir 
        install.install_purelib = self.temp_install_dir 
        install.install_data    = self.temp_install_dir 
        install.install_lib     = self.temp_install_dir 
        install.install_headers = self.temp_install_dir 
        install.run()

        # Horrible trick to reload nipy with our temporary instal
        for key in sys.modules.keys():
            if key.startswith('nipy'):
                sys.modules.pop(key, None)
        sys.path.append(os.path.abspath(self.temp_install_dir))
        # Pop the cwd
        sys.path.pop(0)
        import nipy

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


################################################################################
# Distutils Command class for API generation 
class APIDocs(TempInstall):
    description = \
    """generate API docs """

    user_options = [
        ('None', None, 'this command has no options'),
        ]


    def run(self):
        # First build the project and install it to a temporary location.
        TempInstall.run(self)
        os.chdir('doc')
        try:
            # We are running the API-building script via an
            # system call, but overriding the import path.
            toolsdir = os.path.abspath(pjoin('..', 'tools'))
            build_templates = pjoin(toolsdir, 'build_modref_templates.py')
            cmd = """%s -c 'import sys; sys.path.append("%s"); sys.path.append("%s"); execfile("%s", dict(__name__="__main__"))'""" \
                % (sys.executable, 
                   toolsdir,
                   self.temp_install_dir,
                   build_templates)
            os.system(cmd)
        finally:
            os.chdir('..')


################################################################################
# Code to copy the sphinx-generated html docs in the distribution.
def relative_path(filename):
    """ Return the relative path to the file, assuming the file is
        in the DOC_BUILD_DIR directory.
    """
    length = len(os.path.abspath(DOC_BUILD_DIR)) + 1
    return os.path.abspath(filename)[length:]



################################################################################
# Distutils Command class to clean
class Clean(clean):

    def run(self):
        clean.run(self)
        api_path = os.path.join('doc', 'api', 'generated')
        if os.path.exists(api_path):
            print("Removing %s" % api_path)
            shutil.rmtree(api_path)
        if os.path.exists(DOC_BUILD_DIR):
            print("Removing %s" % DOC_BUILD_DIR)
            shutil.rmtree(DOC_BUILD_DIR)


################################################################################
# Distutils Command class build the docs
if have_sphinx:
    class MyBuildDoc(BuildDoc):
        """ Sub-class the standard sphinx documentation building system, to
            add logics for API generation and matplotlib's plot directive.
        """

        def run(self):
            self.run_command('api_docs')
            # We need to be in the doc directory for to plot_directive
            # and API generation to work
            os.chdir('doc')
            try:
                BuildDoc.run(self)
            finally:
                os.chdir('..')
            self.zip_docs()

        def zip_docs(self):
            if not os.path.exists(DOC_BUILD_DIR):
                raise OSError('Doc directory does not exist.')
            target_file = os.path.join('doc', 'documentation.zip')
            # ZIP_DEFLATED actually compresses the archive. However, there
            # will be a RuntimeError if zlib is not installed, so we check
            # for it. ZIP_STORED produces an uncompressed zip, but does not
            # require zlib.
            try:
                zf = zipfile.ZipFile(target_file, 'w', 
                                            compression=zipfile.ZIP_DEFLATED)
            except RuntimeError:
                warnings.warn('zlib not installed, storing the docs '
                                'without compression')
                zf = zipfile.ZipFile(target_file, 'w', 
                                            compression=zipfile.ZIP_STORED)    

            for root, dirs, files in os.walk(DOC_BUILD_DIR):
                relative = relative_path(root)
                if not relative.startswith('.doctrees'):
                    for f in files:
                        zf.write(os.path.join(root, f), 
                                os.path.join(relative, 'html_docs', f))
            zf.close()

        def finalize_options(self):
            """ Override the default for the documentation build
                directory.
            """
            self.build_dir = os.path.join(*DOC_BUILD_DIR.split(os.sep)[:-1])
            BuildDoc.finalize_options(self)

else: # failed Sphinx import
    # Raise an error when trying to build docs
    class MyBuildDoc(Command):
        user_options = []
        def run(self):
            raise ImportError(
                "Sphinx is not installed, docs cannot be built")
        def initialize_options(self):
            pass
        def finalize_options(self):
            pass


def have_good_cython():
    try:
        from Cython.Compiler.Version import version
    except ImportError:
        return False
    return LooseVersion(version) >= LooseVersion(CYTHON_MIN_VERSION)


def generate_a_pyrex_source(self, base, ext_name, source, extension):
    ''' Monkey patch for numpy build_src.build_src method

    Uses Cython instead of Pyrex.
    '''
    good_cython = have_good_cython()
    if self.inplace or not good_cython:
        target_dir = dirname(base)
    else:
        target_dir = appendpath(self.build_src, dirname(base))
    target_file = pjoin(target_dir, ext_name + '.c')
    depends = [source] + extension.depends
    sources_changed = newer_group(depends, target_file, 'newer') 
    if self.force or sources_changed:
        if good_cython:
            # add distribution (package-wide) include directories, in order
            # to pick up needed .pxd files for cython compilation
            incl_dirs = extension.include_dirs[:]
            dist_incl_dirs = self.distribution.include_dirs
            if dist_incl_dirs is not None:
                incl_dirs += dist_incl_dirs
            import Cython.Compiler.Main
            log.info("cythonc:> %s" % (target_file))
            self.mkpath(target_dir)
            options = Cython.Compiler.Main.CompilationOptions(
                defaults=Cython.Compiler.Main.default_options,
                include_path=incl_dirs,
                output_file=target_file)
            cython_result = Cython.Compiler.Main.compile(source,
                                                       options=options)
            if cython_result.num_errors != 0:
                raise DistutilsError("%d errors while compiling "
                                     "%r with Cython"
                                     % (cython_result.num_errors, source))
        elif sources_changed and os.path.isfile(target_file):
            raise DistutilsError("Cython >=%s required for compiling %r"
                                 " because sources (%s) have changed" %
                                 (CYTHON_MIN_VERSION, source, ','.join(depends)))
        else:
            raise DistutilsError("Cython >=%s required for compiling %r"
                                 " but not available" %
                                 (CYTHON_MIN_VERSION, source))
    return target_file


# Copied from soon-to-be retired nibabel/nisext/sexts.py
# We'll probably drop this way of doing versioning soon
def get_comrec_build(pkg_dir, build_cmd=build_py):
    """ Return extended build command class for recording commit

    The extended command tries to run git to find the current commit, getting
    the empty string if it fails.  It then writes the commit hash into a file
    in the `pkg_dir` path, named ``COMMIT_INFO.txt``.

    In due course this information can be used by the package after it is
    installed, to tell you what commit it was installed from if known.

    To make use of this system, you need a package with a COMMIT_INFO.txt file -
    e.g. ``myproject/COMMIT_INFO.txt`` - that might well look like this::

        # This is an ini file that may contain information about the code state
        [commit hash]
        # The line below may contain a valid hash if it has been substituted during 'git archive'
        archive_subst_hash=$Format:%h$
        # This line may be modified by the install process
        install_hash=

    The COMMIT_INFO file above is also designed to be used with git substitution
    - so you probably also want a ``.gitattributes`` file in the root directory
    of your working tree that contains something like this::

       myproject/COMMIT_INFO.txt export-subst

    That will cause the ``COMMIT_INFO.txt`` file to get filled in by ``git
    archive`` - useful in case someone makes such an archive - for example with
    via the github 'download source' button.

    Although all the above will work as is, you might consider having something
    like a ``get_info()`` function in your package to display the commit
    information at the terminal.  See the ``pkg_info.py`` module in the nipy
    package for an example.
    """
    class MyBuildPy(build_cmd):
        ''' Subclass to write commit data into installation tree '''
        def run(self):
            build_cmd.run(self)
            import subprocess
            proc = subprocess.Popen('git rev-parse --short HEAD',
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    shell=True)
            repo_commit, _ = proc.communicate()
            # Fix for python 3
            repo_commit = str(repo_commit)
            # We write the installation commit even if it's empty
            cfg_parser = ConfigParser()
            cfg_parser.read(pjoin(pkg_dir, 'COMMIT_INFO.txt'))
            cfg_parser.set('commit hash', 'install_hash', repo_commit)
            out_pth = pjoin(self.build_lib, pkg_dir, 'COMMIT_INFO.txt')
            cfg_parser.write(open(out_pth, 'wt'))
    return MyBuildPy


# The command classes for distutils, used by setup.py
cmdclass = {'api_docs': APIDocs,
            'clean': Clean,
            'build_sphinx': MyBuildDoc}



""" Testing interpolation module
"""

from itertools import product

import numpy as np

from scipy.ndimage import map_coordinates

from nipy.core.api import Image, vox2mni

from ..interpolation import ImageInterpolator

from numpy.testing import (assert_almost_equal,
                           assert_array_equal)

from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)


def test_interp_obj():
    arr = np.arange(24).reshape((2, 3, 4))
    coordmap =  vox2mni(np.eye(4))
    img = Image(arr, coordmap)
    interp = ImageInterpolator(img)
    assert_equal(interp.mode, 'constant')
    assert_equal(interp.order, 3)
    # order is read-only
    assert_raises(AttributeError,
                  setattr,
                  interp,
                  'order',
                  1)
    interp = ImageInterpolator(img, mode='nearest')
    assert_equal(interp.mode, 'nearest')
    # mode is read-only
    assert_raises(AttributeError,
                  setattr,
                  interp,
                  'mode',
                  'reflect')


def test_interpolator():
    shape = (2, 3, 4)
    arr = np.arange(24).reshape(shape)
    coordmap =  vox2mni(np.eye(4))
    img = Image(arr, coordmap)
    ixs = np.indices(arr.shape).astype(float)
    for order in range(5):
        interp = ImageInterpolator(img, mode='nearest', order=order)
        # Interpolate at existing points.
        assert_almost_equal(interp.evaluate(ixs), arr)
        # Interpolate at half voxel shift
        ixs_x_shift = ixs.copy()
        # Interpolate inside and outside at knots
        ixs_x_shift[0] += 1
        res = interp.evaluate(ixs_x_shift)
        assert_almost_equal(res, np.tile(arr[1], (2, 1, 1)))
        ixs_x_shift[0] -= 2
        res = interp.evaluate(ixs_x_shift)
        assert_almost_equal(res, np.tile(arr[0], (2, 1, 1)))
        # Interpolate at mid-points inside and outside
        ixs_x_shift[0] += 0.5
        res = interp.evaluate(ixs_x_shift)
        # Check inside.
        mid_arr = np.mean(arr, axis=0) if order > 0 else arr[1]
        assert_almost_equal(res[1], mid_arr)
        # Interpolate off top right corner with different modes
        assert_almost_equal(interp.evaluate([0, 0, 4]), arr[0, 0, -1])
        interp = ImageInterpolator(img, mode='constant', order=order, cval=0)
        assert_array_equal(interp.evaluate([0, 0, 4]), 0)
        interp = ImageInterpolator(img, mode='constant', order=order, cval=1)
        assert_array_equal(interp.evaluate([0, 0, 4]), 1)
        # Check against direct ndimage interpolation
        # Need floating point input array to replicate
        # our floating point backing store.
        farr = arr.astype(float)
        for offset, axis, mode in product(np.linspace(-2, 2, 15),
                                          range(3),
                                          ('nearest', 'constant')):
            interp = ImageInterpolator(img, mode=mode, order=order)
            coords = ixs.copy()
            slicer = tuple(None if i == axis else 0 for i in range(3))
            coords[slicer] = coords[slicer] + offset
            actual = interp.evaluate(coords)
            expected = map_coordinates(farr, coords, mode=mode, order=order)
            assert_almost_equal(actual, expected)
            del interp



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Generic algorithms such as registration, statistics, simulation, etc.
"""
from __future__ import absolute_import
__docformat__ = 'restructuredtext'

from . import statistics
from . import fwhm, interpolation, kernel_smooth, diagnostics

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



from __future__ import absolute_import

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('formula',parent_package,top_path)
    config.add_subpackage('tests')
    return config

if __name__ == '__main__':

    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This figure is meant to represent the neuronal event-related model and a
coefficient of +1 for Faces, -2 for Objects.
"""

import numpy as np

import matplotlib.pyplot as plt

from sympy import Symbol, Heaviside, lambdify

ta = [0, 4, 8, 12, 16]
tb = [2, 6, 10, 14, 18]
ba = Symbol('ba')
bb = Symbol('bb')
t = Symbol('t')
fa = sum([Heaviside(t - _t) for _t in ta]) * ba
fb = sum([Heaviside(t - _t) for _t in tb]) * bb
N = fa + fb

Nn = N.subs(ba, 1)
Nn = Nn.subs(bb, -2)

# Use Numpy heaviside for lambdify, with y=1 for x=0.
modules = [{'Heaviside': lambda x: np.heaviside(x, 1)}, 'numpy']
neuronal_func = lambdify(t, Nn, modules=modules)

tt = np.linspace(-1, 21, 1201)
neuronal = neuronal_func(tt)

plt.step(tt, neuronal)

a = plt.gca()
a.set_ylim([-5.5, 1.5])
a.set_ylabel('Neuronal (cumulative)')
a.set_xlabel('Time')

plt.show()



#!/usr/bin/env python3
""" Refresh README.rst file from long description

Should be run from project root (containing setup.py)
"""
from __future__ import print_function

import os
import sys


def main():
    project_name = sys.argv[1]
    readme_lines = []
    with open('README.rst', 'rt') as fobj:
        for line in fobj:
            readme_lines.append(line)
            if line.startswith('.. Following contents should be'):
                break
        else:
            raise ValueError('Expected comment not found')

    rel = {}
    with open(os.path.join(project_name, 'info.py'), 'rt') as fobj:
        exec(fobj.read(), {}, rel)

    readme = ''.join(readme_lines) + '\n' + rel['LONG_DESCRIPTION']

    with open('README.rst', 'wt') as fobj:
        fobj.write(readme)

    print('Done')

if __name__ == '__main__':
    main()



from __future__ import absolute_import

import numpy as np

from scipy.stats import norm
import scipy.linalg as spl


from ..utils import (multiple_mahalanobis, z_score, multiple_fast_inv,
                     check_cast_bin8)
from nose.tools import assert_true, assert_equal, assert_raises
from numpy.testing import (assert_almost_equal, assert_array_almost_equal,
                           assert_array_equal)


def test_z_score():
    p = np.random.rand(10)
    z = z_score(p)
    assert_array_almost_equal(norm.sf(z), p)


def test_mahalanobis():
    x = np.random.rand(100) / 100
    A = np.random.rand(100, 100) / 100
    A = np.dot(A.transpose(), A) + np.eye(100)
    mah = np.dot(x, np.dot(np.linalg.inv(A), x))
    assert_almost_equal(mah, multiple_mahalanobis(x, A), decimal=1)


def test_mahalanobis2():
    x = np.random.randn(100, 3)
    Aa = np.zeros([100, 100, 3])
    for i in range(3):
        A = np.random.randn(120, 100)
        A = np.dot(A.T, A)
        Aa[:, :, i] = A
    i = np.random.randint(3)
    mah = np.dot(x[:, i], np.dot(np.linalg.inv(Aa[:, :, i]), x[:, i]))
    f_mah = (multiple_mahalanobis(x, Aa))[i]
    assert_almost_equal(mah, f_mah)


def test_multiple_fast_inv():
    shape = (10, 20, 20)
    X = np.random.randn(*shape)
    X_inv_ref = np.zeros(shape)
    for i in range(shape[0]):
        X[i] = np.dot(X[i], X[i].T)
        X_inv_ref[i] = spl.inv(X[i])
    X_inv = multiple_fast_inv(X)
    assert_array_almost_equal(X_inv_ref, X_inv)


def assert_equal_bin8(actual, expected):
    res = check_cast_bin8(actual)
    assert_equal(res.shape, actual.shape)
    assert_true(res.dtype.type == np.uint8)
    assert_array_equal(res, expected)


def test_check_cast_bin8():
    # Function to return np.uint8 array with check whether array is binary.
    for in_dtype in np.sctypes['int'] + np.sctypes['uint']:
        assert_equal_bin8(np.array([0, 1, 1, 1], in_dtype), [0, 1, 1, 1])
        assert_equal_bin8(np.array([[0, 1], [1, 1]], in_dtype),
                          [[0, 1], [1, 1]])
        assert_raises(ValueError, check_cast_bin8,
                      np.array([0, 1, 2], dtype=in_dtype))
    for in_dtype in np.sctypes['float']:
        assert_equal_bin8(np.array([0, 1, 1, -0], np.float), [0, 1, 1, 0])
        assert_equal_bin8(np.array([[0, 1], [1, -0]], np.float),
                          [[0, 1], [1, 0]])
        assert_raises(ValueError, check_cast_bin8,
                      np.array([0, 0.1, 1], dtype=in_dtype))
        assert_raises(ValueError, check_cast_bin8,
                      np.array([0, -1, 1], dtype=in_dtype))



#!/usr/bin/env python3
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
Example showing how to use the parcel generator.

We load an image with ROI definitions and calculate the number of voxels in each
ROI.
"""
print(__doc__)

from os.path import dirname, join as pjoin

import nipy

from nipy.core.utils.generators import parcels

OUR_PATH = dirname(__file__)
DATA_PATH = pjoin(OUR_PATH, '..', 'data')
BG_IMAGE_FNAME = pjoin(DATA_PATH, 'mni_basal_ganglia.nii.gz')

bg_img = nipy.load_image(BG_IMAGE_FNAME)
bg_data = bg_img.get_data()

"""
I happen to know that the image has these codes:

14 - Left striatum
16 - Right striatum
39 - Left caudate
53 - Right caudate

All the other voxels are zero, I don't want those.
"""

print("Number of voxels for L, R striatum; L, R caudate")
for mask in parcels(bg_data, exclude=(0,)):
    print(mask.sum())

""" Given we know the codes we can also give them directly """
print("Again with the number of voxels for L, R striatum; L, R caudate")
for mask in parcels(bg_data, labels=(14, 16, 39, 53)):
    print(mask.sum())



#!/usr/bin/env python3
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
''' Single subject analysis script for SPM / Open FMRI ds105

https://openfmri.org/dataset/ds000105

Download and extract the ds105 archive to some directory.

Run this script with::

    process_ds105.py ~/data/ds105

where ``~/data/ds105`` is the directory containing the ds105 data.

The example uses the very basic MATLAB / SPM interface routines in NIPY.

If you need more than very basic use, please consider using nipype.  nipype has
extended capabilities to interface with external tools and for dataflow
management. nipype can handle vanilla SPM in MATLAB or SPM run through the
MATLAB common runtime (free from MATLAB Licensing).
'''
from __future__ import print_function, division, absolute_import

import sys
from copy import deepcopy
from os.path import join as pjoin, abspath, splitext, isfile
from glob import glob
from warnings import warn
import gzip

import numpy as np

import nipy.interfaces.matlab as nimat
from nipy.interfaces.spm import (spm_info, make_job, scans_for_fnames,
                                 run_jobdef, fnames_presuffix, fname_presuffix,
                                 fltcols)


# The batch scripts currently need SPM5
nimat.matlab_cmd = 'matlab-spm8 -nodesktop -nosplash'

# This definition is partly for slice timing. We can't do slice timing for this
# dataset because the slice dimension is the first, and SPM assumes it is the
# last.
N_SLICES = 40 # X slices
STUDY_DEF = dict(
    TR = 2.5,
    n_slices = N_SLICES,
    time_to_space = (list(range(1, N_SLICES, 2)) +
                     list(range(2, N_SLICES, 2)))
)


def _sorted_prefer_nii(file_list):
    """ Strip any filanames ending nii.gz if matching .nii filename in list
    """
    preferred = []
    for fname in file_list:
        if not fname.endswith('.gz'):
            preferred.append(fname)
        else:
            nogz, ext = splitext(fname)
            if not nogz in file_list:
                preferred.append(fname)
    return sorted(preferred)


def get_data(data_path, subj_id):
    data_path = abspath(data_path)
    data_def = {}
    subject_path = pjoin(data_path, 'sub%03d' % subj_id)
    functionals = _sorted_prefer_nii(
        glob(pjoin(subject_path, 'BOLD', 'task*', 'bold*.nii*')))
    anatomicals = _sorted_prefer_nii(
        glob(pjoin(subject_path, 'anatomy', 'highres001.nii*')))
    for flist in (anatomicals, functionals):
        for i, fname in enumerate(flist):
            nogz, gz_ext = splitext(fname)
            if gz_ext == '.gz':
                if not isfile(nogz):
                    contents = gzip.open(fname, 'rb').read()
                    with open(nogz, 'wb') as fobj:
                        fobj.write(contents)
                flist[i] = nogz
    if len(anatomicals) == 0:
        data_def['anatomical'] = None
    else:
        data_def['anatomical'] = anatomicals[0]
    data_def['functionals'] = functionals
    return data_def


def default_ta(tr, nslices):
    slice_time = tr / float(nslices)
    return slice_time * (nslices - 1)


class SPMSubjectAnalysis(object):
    """ Class to preprocess single subject in SPM
    """
    def __init__(self, data_def, study_def, ana_def):
        self.data_def = deepcopy(data_def)
        self.study_def = self.add_study_defaults(study_def)
        self.ana_def = self.add_ana_defaults(deepcopy(ana_def))

    def add_study_defaults(self, study_def):
        full_study_def = deepcopy(study_def)
        if 'TA' not in full_study_def:
            full_study_def['TA'] = default_ta(
                full_study_def['TR'], full_study_def['n_slices'])
        return full_study_def

    def add_ana_defaults(self, ana_def):
        full_ana_def = deepcopy(ana_def)
        if 'fwhm' not in full_ana_def:
            full_ana_def['fwhm'] = 8.0
        return full_ana_def

    def slicetime(self, in_prefix='', out_prefix='a'):
        sess_scans = scans_for_fnames(
            fnames_presuffix(self.data_def['functionals'], in_prefix))
        sdef = self.study_def
        stinfo = make_job('temporal', 'st', {
                'scans': sess_scans,
                'so': sdef['time_to_space'],
                'tr': sdef['TR'],
                'ta': sdef['TA'],
                'nslices': float(sdef['n_slices']),
                'refslice':1,
                'prefix': out_prefix,
                })
        run_jobdef(stinfo)
        return out_prefix + in_prefix


    def realign(self, in_prefix=''):
        sess_scans = scans_for_fnames(
            fnames_presuffix(self.data_def['functionals'], in_prefix))
        rinfo = make_job('spatial', 'realign', [{
                'estimate':{
                    'data':sess_scans,
                    'eoptions':{
                        'quality': 0.9,
                        'sep': 4.0,
                        'fwhm': 5.0,
                        'rtm': True,
                        'interp': 2.0,
                        'wrap': [0.0,0.0,0.0],
                        'weight': []
                        }
                    }
                }])
        run_jobdef(rinfo)
        return in_prefix

    def reslice(self, in_prefix='', out_prefix='r', out=('1..n', 'mean')):
        which = [0, 0]
        if 'mean' in out:
            which[1] = 1
        if '1..n' in out or 'all' in out:
            which[0] = 2
        elif '2..n' in out:
            which[0] = 1
        sess_scans = scans_for_fnames(
            fnames_presuffix(self.data_def['functionals'], in_prefix))
        rsinfo = make_job('spatial', 'realign', [{
                'write':{
                    'data': np.vstack(sess_scans.flat),
                    'roptions':{
                        'which': which,
                        'interp':4.0,
                        'wrap':[0.0,0.0,0.0],
                        'mask':True,
                        'prefix': out_prefix
                        }
                    }
                }])
        run_jobdef(rsinfo)
        return out_prefix + in_prefix

    def coregister(self, in_prefix=''):
        func1 = self.data_def['functionals'][0]
        mean_fname = fname_presuffix(func1, 'mean' + in_prefix)
        crinfo = make_job('spatial', 'coreg', [{
                'estimate':{
                    'ref': np.asarray(mean_fname, dtype=object),
                    'source': np.asarray(self.data_def['anatomical'],
                                         dtype=object),
                    'other': [''],
                    'eoptions':{
                        'cost_fun':'nmi',
                        'sep':[4.0, 2.0],
                        'tol':np.array(
                                [0.02,0.02,0.02,
                                0.001,0.001,0.001,
                                0.01,0.01,0.01,
                                0.001,0.001,0.001]).reshape(1,12),
                        'fwhm':[7.0, 7.0]
                        }
                    }
                }])
        run_jobdef(crinfo)
        return in_prefix

    def seg_norm(self, in_prefix=''):
        def_tpms = np.zeros((3,1), dtype=object)
        spm_path = spm_info.spm_path
        def_tpms[0] = pjoin(spm_path, 'tpm', 'grey.nii'),
        def_tpms[1] = pjoin(spm_path, 'tpm', 'white.nii'),
        def_tpms[2] = pjoin(spm_path, 'tpm', 'csf.nii')
        data = np.zeros((1,), dtype=object)
        data[0] = self.data_def['anatomical']
        sninfo = make_job('spatial', 'preproc', {
                'data': data,
                'output':{
                    'GM':fltcols([0,0,1]),
                    'WM':fltcols([0,0,1]),
                    'CSF':fltcols([0,0,0]),
                    'biascor':1.0,
                    'cleanup':False,
                    },
                'opts':{
                    'tpm':def_tpms,
                    'ngaus':fltcols([2,2,2,4]),
                    'regtype':'mni',
                    'warpreg':1.0,
                    'warpco':25.0,
                    'biasreg':0.0001,
                    'biasfwhm':60.0,
                    'samp':3.0,
                    'msk':np.array([], dtype=object),
                    }
                })
        run_jobdef(sninfo)
        return in_prefix

    def norm_write(self, in_prefix='', out_prefix='w'):
        sess_scans = scans_for_fnames(
            fnames_presuffix(self.data_def['functionals'], in_prefix))
        matname = fname_presuffix(self.data_def['anatomical'],
                                suffix='_seg_sn.mat',
                                use_ext=False)
        subj = {
            'matname': np.zeros((1,), dtype=object),
            'resample': np.vstack(sess_scans.flat),
            }
        subj['matname'][0] = matname
        roptions = {
            'preserve':False,
            'bb':np.array([[-78,-112, -50],[78,76,85.0]]),
            'vox':fltcols([2.0,2.0,2.0]),
            'interp':1.0,
            'wrap':[0.0,0.0,0.0],
            'prefix': out_prefix,
            }
        nwinfo = make_job('spatial', 'normalise', [{
                'write':{
                    'subj': subj,
                    'roptions': roptions,
                    }
                }])
        run_jobdef(nwinfo)
        # knock out the list of images, replacing with only one
        subj['resample'] = np.zeros((1,), dtype=object)
        subj['resample'][0] = self.data_def['anatomical']
        roptions['interp'] = 4.0
        run_jobdef(nwinfo)
        return out_prefix + in_prefix

    def smooth(self, in_prefix='', out_prefix='s'):
        fwhm = self.ana_def['fwhm']
        try:
            len(fwhm)
        except TypeError:
            fwhm = [fwhm] * 3
        fwhm = np.asarray(fwhm, dtype=np.float).reshape(1,3)
        sess_scans = scans_for_fnames(
            fnames_presuffix(self.data_def['functionals'], in_prefix))
        sinfo = make_job('spatial', 'smooth',
                        {'data':np.vstack(sess_scans.flat),
                        'fwhm':fwhm,
                        'dtype':0})
        run_jobdef(sinfo)
        return out_prefix + in_prefix


def process_subject(ddef, study_def, ana_def):
    """ Process subject from subject data dict `ddef`
    """
    if not ddef['anatomical']:
        warn("No anatomical, aborting processing")
        return
    ana = SPMSubjectAnalysis(ddef, study_def, ana_def)
    # st_prefix = ana.slicetime('') # We can't run slice timing
    st_prefix = ''
    ana.realign(in_prefix=st_prefix)
    ana.reslice(in_prefix=st_prefix, out=('mean',))
    ana.coregister(in_prefix=st_prefix)
    ana.seg_norm()
    n_st_prefix = ana.norm_write(st_prefix)
    ana.smooth(n_st_prefix)


def get_subjects(data_path, subj_ids, study_def, ana_def):
    ddefs = []
    for subj_id in subj_ids:
        ddefs.append(get_data(data_path, subj_id))
    return ddefs


def main():
    try:
        data_path = sys.argv[1]
    except IndexError:
        raise OSError('Need ds105 data path as input')
    if len(sys.argv) > 2:
        subj_ids = [int(id) for id in sys.argv[2:]]
    else:
        subj_ids = range(1, 7)
    for subj_id in subj_ids:
        ddef = get_data(data_path, subj_id)
        assert len(ddef['functionals']) in (11, 12)
        process_subject(ddef, STUDY_DEF, {})


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
""" Fix sphinx latex output for longtable
"""
import re
import sys
import codecs

lt_LL = re.compile(
    r"longtable}{(L+)}")

def replacer(match):
    args =  '|' + 'l|' * len(match.groups()[0])
    return "longtable}{%s}" % args


if len(sys.argv) != 2:
    raise RuntimeError("Enter path to tex file only")
file_path = sys.argv[1]

with codecs.open(file_path, 'r', encoding='utf8') as fobj:
    unfixed_tex = fobj.readlines()
with codecs.open(file_path, 'w', encoding='utf8') as fobj:
    for line in unfixed_tex:
        line = lt_LL.sub(replacer, line, 1)
        fobj.write(line)



"""
The estimators for the intrinsic volumes appearing in this module
were partially supported by NSF grant DMS-0405970.

Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
   with an application to brain mapping."
   Journal of the American Statistical Association, 102(479):913-928.

"""
cimport cython

import numpy as np
cimport numpy as np

from scipy.sparse import dok_matrix

# Array helper
from nipy.utils.arrays import strides_from

# local imports
from .utils import cube_with_strides_center, join_complexes, check_cast_bin8


cdef double PI = np.pi


cdef extern from "math.h" nogil:
    double floor(double x)
    double sqrt(double x)
    double fabs(double x)
    double log2(double x)
    double acos(double x)
    bint isnan(double x)


cpdef double mu3_tet(double D00, double D01, double D02, double D03,
                     double D11, double D12, double D13,
                     double D22, double D23,
                     double D33) nogil:
  """ Compute the 3rd intrinsic volume of a tetrahedron.

  3rd intrinsic volume (just volume in this case) of a tetrahedron with
  coordinates implied by dot products below.

  Parameters
  ----------
  D00 : float
    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is
    ``cv0.dot(cv0)``
  D01 : float
    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second
    vertex.
  D02 : float
    ``cv0.dot(cv2)``
  D03 : float
    ``cv0.dot(cv3)``
  D11 : float
    ``cv1.dot(cv1)``
  D12 : float
    ``cv1.dot(cv2)``
  D13 : float
    ``cv1.dot(cv3)``
  D22 : float
    ``cv2.dot(cv2)``
  D23 : float
    ``cv2.dot(cv2)``
  D33 : float
    ``cv3.dot(cv3)``

  Returns
  -------
  mu3 : float
    volume of tetrahedron
  """
  cdef double C00, C01, C02, C11, C12, C22, v2
  C00 = D00 - 2*D03 + D33
  C01 = D01 - D13 - D03 + D33
  C02 = D02 - D23 - D03 + D33
  C11 = D11 - 2*D13 + D33
  C12 = D12 - D13 - D23 + D33
  C22 = D22 - 2*D23 + D33
  v2 = (C00 * (C11 * C22 - C12 * C12) -
        C01 * (C01 * C22 - C02 * C12) +
        C02 * (C01 * C12 - C11 * C02))
  # Rounding errors near 0 cause NaNs
  if v2 <= 0:
      return 0
  return sqrt(v2) / 6.


cpdef double mu2_tet(double D00, double D01, double D02, double D03,
                     double D11, double D12, double D13,
                     double D22, double D23,
                     double D33) nogil:
  """ Compute the 2nd intrinsic volume of tetrahedron

  2nd intrinsic volume (half the surface area) of a tetrahedron with coordinates
  implied by dot products below.

  Parameters
  ----------
  D00 : float
    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is
    ``cv0.dot(cv0)``
  D01 : float
    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second
    vertex.
  D02 : float
    ``cv0.dot(cv2)``
  D03 : float
    ``cv0.dot(cv3)``
  D11 : float
    ``cv1.dot(cv1)``
  D12 : float
    ``cv1.dot(cv2)``
  D13 : float
    ``cv1.dot(cv3)``
  D22 : float
    ``cv2.dot(cv2)``
  D23 : float
    ``cv2.dot(cv2)``
  D33 : float
    ``cv3.dot(cv3)``

  Returns
  -------
  mu2 : float
    Half tetrahedron surface area
  """
  cdef double mu = 0
  mu += mu2_tri(D00, D01, D02, D11, D12, D22)
  mu += mu2_tri(D00, D02, D03, D22, D23, D33)
  mu += mu2_tri(D11, D12, D13, D22, D23, D33)
  mu += mu2_tri(D00, D01, D03, D11, D13, D33)
  return mu * 0.5


cpdef double mu1_tet(double D00, double D01, double D02, double D03,
                     double D11, double D12, double D13,
                     double D22, double D23,
                     double D33) nogil:
  """ Return 3rd intrinsic volume of tetrahedron

  Compute the 3rd intrinsic volume (sum of external angles * edge
  lengths) of a tetrahedron for which the input arguments represent the
  coordinate dot products of the vertices.

  Parameters
  ----------
  D00 : float
    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is
    ``cv0.dot(cv0)``
  D01 : float
    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second
    vertex.
  D02 : float
    ``cv0.dot(cv2)``
  D03 : float
    ``cv0.dot(cv3)``
  D11 : float
    ``cv1.dot(cv1)``
  D12 : float
    ``cv1.dot(cv2)``
  D13 : float
    ``cv1.dot(cv3)``
  D22 : float
    ``cv2.dot(cv2)``
  D23 : float
    ``cv2.dot(cv2)``
  D33 : float
    ``cv3.dot(cv3)``

  Returns
  -------
  mu1 : float
    3rd intrinsic volume of tetrahedron
  """
  cdef double mu
  mu = 0
  mu += _mu1_tetface(D00, D01, D11, D02, D03, D12, D13, D22, D23, D33)
  mu += _mu1_tetface(D00, D02, D22, D01, D03, D12, D23, D11, D13, D33)
  mu += _mu1_tetface(D00, D03, D33, D01, D02, D13, D23, D11, D12, D22)
  mu += _mu1_tetface(D11, D12, D22, D01, D13, D02, D23, D00, D03, D33)
  mu += _mu1_tetface(D11, D13, D33, D01, D12, D03, D23, D00, D02, D22)
  mu += _mu1_tetface(D22, D23, D33, D02, D12, D03, D13, D00, D01, D11)
  return mu


cdef inline double limited_acos(double val) nogil:
    """ Check for -1 <= val <= 1 before returning acos(val)

    Avoids nan values from small rounding errors
    """
    if val >= 1:
        return 0
    elif val <= -1:
        return PI
    return acos(val)


@cython.cdivision(True)
cpdef double _mu1_tetface(double Ds0s0,
                          double Ds0s1,
                          double Ds1s1,
                          double Ds0t0,
                          double Ds0t1,
                          double Ds1t0,
                          double Ds1t1,
                          double Dt0t0,
                          double Dt0t1,
                          double Dt1t1) nogil:
    cdef double A00, A01, A02, A11, A12, A22, np_len, a, acosval
    cdef double length, norm_proj0, norm_proj1, inner_prod_proj

    A00 = Ds1s1 - 2 * Ds0s1 + Ds0s0
    # all norms divided by this value, leading to NaN value for output, for
    # values <= 0
    if A00 <= 0:
      return 0
    A11 = Dt0t0 - 2 * Ds0t0 + Ds0s0
    A22 = Dt1t1 - 2 * Ds0t1 + Ds0s0
    A01 = Ds1t0 - Ds0t0 - Ds0s1 + Ds0s0
    A02 = Ds1t1 - Ds0t1 - Ds0s1 + Ds0s0
    A12 = Dt0t1 - Ds0t0 - Ds0t1 + Ds0s0
    length = sqrt(A00)
    norm_proj0 = A11 - A01 * A01 / A00
    norm_proj1 = A22 - A02 * A02 / A00
    inner_prod_proj = A12 - A01 * A02 / A00
    np_len = norm_proj0 * norm_proj1
    if np_len <= 0: # would otherwise lead to NaN return value
        return 0
    # hedge for small rounding errors above 1 and below -1
    acosval = limited_acos(inner_prod_proj / sqrt(np_len))
    a = (PI - acosval) * length / (2 * PI)
    return a


cpdef double mu2_tri(double D00, double D01, double D02,
                     double D11, double D12,
                     double D22) nogil:
  """ Compute the 2nd intrinsic volume of triangle

  2nd intrinsic volume (just area in this case) of a triangle with coordinates
  implied by the dot products below.

  Parameters
  ----------
  D00 : float
    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is
    ``cv0.dot(cv0)``
  D01 : float
    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second
    vertex.
  D02 : float
    ``cv0.dot(cv2)``
  D11 : float
    ``cv1.dot(cv1)``
  D12 : float
    ``cv1.dot(cv2)``
  D22 : float
    ``cv2.dot(cv2)``

  Returns
  -------
  mu2 : float
    area of triangle
  """
  cdef double C00, C01, C11, L
  C00 = D11 - 2*D01 + D00
  C01 = D12 - D01 - D02 + D00
  C11 = D22 - 2*D02 + D00
  L = C00 * C11 - C01 * C01
  # Negative area appeared to result from floating point errors on PPC
  if L < 0:
      return 0.0
  return sqrt(L) * 0.5


cpdef double mu1_tri(double D00, double D01, double D02,
                     double D11, double D12,
                     double D22) nogil:
  """ Compute the 1st intrinsic volume of triangle

  1st intrinsic volume (1/2 the perimeter) of a triangle with coordinates
  implied by the dot products below.

  Parameters
  ----------
  D00 : float
    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is
    ``cv0.dot(cv0)``
  D01 : float
    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second
    vertex.
  D02 : float
    ``cv0.dot(cv2)``
  D11 : float
    ``cv1.dot(cv1)``
  D12 : float
    ``cv1.dot(cv2)``
  D22 : float
    ``cv2.dot(cv2)``

  Returns
  -------
  mu1 : float
    1/2 perimeter of triangle
  """
  cdef double mu = 0
  mu += mu1_edge(D00, D01, D11)
  mu += mu1_edge(D00, D02, D22)
  mu += mu1_edge(D11, D12, D22)
  return mu * 0.5


cpdef double mu1_edge(double D00, double D01, double D11) nogil:
  """ Compute the 1st intrinsic volume (length) of line segment

  Length of a line segment with vertex coordinates implied by dot products
  below.

  Parameters
  ----------
  D00 : float
    If ``cv0`` is a 3-vector of coordinates for the line start, `D00` is
    ``cv0.dot(cv0)``
  D01 : float
    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the line end.
  D11 : float
    ``cv1.dot(cv1)``

  Returns
  -------
  mu0 : float
    length of line segment
  """
  return sqrt(D00 - 2*D01 + D11)


def EC3d(mask):
    """ Compute Euler characteristic of region within `mask`

    Given a 3d `mask`, compute the 0th intrinsic volume (Euler characteristic)
    of the masked region. The region is broken up into tetrahedra / triangles /
    edges / vertices, which are included based on whether all voxels in the
    tetrahedron / triangle / edge / vertex are in the mask or not.

    Parameters
    ----------
    mask : ndarray shape (i,j,k)
         Binary mask determining whether or not a voxel is in the mask.

    Returns
    -------
    mu0 : int
        Euler characteristic

    Notes
    -----
    We check whether `mask` is binary.

    The 3d cubes are triangulated into 6 tetrahedra of equal volume, as
    described in the reference below.

    Raises
    ------
    ValueError
        If any value in the mask is outside {0, 1}

    References
    ----------
    Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
      with an application to brain mapping."
      Journal of the American Statistical Association, 102(479):913-928.
    """
    cdef:
        # 'flattened' mask (1d array)
        np.ndarray[np.uint8_t, ndim=1] fpmask
        # d3 and d4 are lists of triangles and tetrahedra
        # associated to particular voxels in the cuve
        np.ndarray[np.intp_t, ndim=2] d2
        np.ndarray[np.intp_t, ndim=2] d3
        np.ndarray[np.intp_t, ndim=2] d4
        # scalars
        np.uint8_t m
        np.npy_intp i, j, k, l, s0, s1, s2, ds2, ds3, ds4, index, nvox
        np.npy_intp ss0, ss1, ss2 # strides
        np.ndarray[np.intp_t, ndim=1] strides
        np.npy_intp v0, v1, v2, v3 # vertices
        np.npy_intp l0 = 0

    pmask_shape = np.array(mask.shape) + 1
    s0, s1, s2 = pmask_shape[:3]
    pmask = np.zeros(pmask_shape, dtype=np.uint8)
    pmask[:-1, :-1, :-1] = check_cast_bin8(mask)
    fpmask = pmask.reshape(-1)

    strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)

    # First do the interior contributions.
    # We first figure out which vertices, edges, triangles, tetrahedra
    # are uniquely associated with an interior voxel
    union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
                             cube_with_strides_center((0,1,0), strides),
                             cube_with_strides_center((0,1,1), strides),
                             cube_with_strides_center((1,0,0), strides),
                             cube_with_strides_center((1,0,1), strides),
                             cube_with_strides_center((1,1,0), strides),
                             cube_with_strides_center((1,1,1), strides)])
    c = cube_with_strides_center((0,0,0), strides)

    d4 = np.array(list(c[4].difference(union[4])))
    d3 = np.array(list(c[3].difference(union[3])))
    d2 = np.array(list(c[2].difference(union[2])))

    ds2 = d2.shape[0]
    ds3 = d3.shape[0]
    ds4 = d4.shape[0]

    ss0 = strides[0]
    ss1 = strides[1]
    ss2 = strides[2]

    nvox = mask.size

    for i in range(s0-1):
        for j in range(s1-1):
            for k in range(s2-1):
                index = i*ss0+j*ss1+k*ss2
                for l in range(ds4):
                    v0 = index + d4[l,0]
                    m = fpmask[v0]
                    if m:
                        v1 = index + d4[l,1]
                        v2 = index + d4[l,2]
                        v3 = index + d4[l,3]
                        m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]
                        l0 = l0 - m

                for l in range(ds3):
                    v0 = index + d3[l,0]
                    m = fpmask[v0]
                    if m:
                        v1 = index + d3[l,1]
                        v2 = index + d3[l,2]
                        m = m * fpmask[v1] * fpmask[v2]
                        l0 = l0 + m

                for l in range(ds2):
                    v0 = index + d2[l,0]
                    m = fpmask[v0]
                    if m:
                        v1 = index + d2[l,1]
                        m = m * fpmask[v1]
                        l0 = l0 - m

    # fpmask has the same sum as mask, but with predictable dtype
    l0 += fpmask.sum()
    return l0


def Lips3d(coords, mask):
    """ Estimated intrinsic volumes within masked region given coordinates

    Given a 3d `mask` and coordinates `coords`, estimate the intrinsic volumes
    of the masked region. The region is broken up into tetrahedra / triangles /
    edges / vertices, which are included based on whether all voxels in the
    tetrahedron / triangle / edge / vertex are in the mask or not.

    Parameters
    ----------
    coords : ndarray shape (N, i, j, k)
         Coordinates for the voxels in the mask. ``N`` will often be 3 (for 3
         dimensional coordinates), but can be any integer > 0
    mask : ndarray shape (i, j, k)
         Binary mask determining whether or not
         a voxel is in the mask.

    Returns
    -------
    mu : ndarray
        Array of intrinsic volumes [mu0, mu1, mu2, mu3], being, respectively:
        #. Euler characteristic
        #. 2 * mean caliper diameter
        #. 0.5 * surface area
        #. Volume.

    Notes
    -----
    We check whether `mask` is binary.

    The 3d cubes are triangulated into 6 tetrahedra of equal volume, as
    described in the reference below.

    Raises
    ------
    ValueError
        If any value in the mask is outside {0, 1}

    References
    ----------
    Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
      with an application to brain mapping."
      Journal of the American Statistical Association, 102(479):913-928.
    """
    if mask.shape != coords.shape[1:]:
        raise ValueError('shape of mask does not match coordinates')
    # if the data can be squeezed, we must use the lower dimensional function
    mask = np.squeeze(mask)
    if mask.ndim < 3:
        value = np.zeros(4)
        coords = coords.reshape((coords.shape[0],) + mask.shape)
        if mask.ndim == 2:
            value[:3] = Lips2d(coords, mask)
        elif mask.ndim == 1:
            value[:2] = Lips1d(coords, mask)
        return value

    cdef:
        # c-level versions of the arrays
        # 'flattened' coords (2d array)
        np.ndarray[np.float_t, ndim=2] fcoords
        np.ndarray[np.float_t, ndim=2] D
        # 'flattened' mask (1d array)
        np.ndarray[np.uint8_t, ndim=1] fmask
        np.ndarray[np.uint8_t, ndim=1] fpmask
        # d3 and d4 are lists of triangles and tetrahedra
        # associated to particular voxels in the cube
        np.ndarray[np.intp_t, ndim=2] d4
        np.ndarray[np.intp_t, ndim=2] m4
        np.ndarray[np.intp_t, ndim=2] d3
        np.ndarray[np.intp_t, ndim=2] m3
        np.ndarray[np.intp_t, ndim=2] d2
        np.ndarray[np.intp_t, ndim=2] m2
        np.ndarray[np.intp_t, ndim=1] cvertices
        # scalars
        np.uint8_t m, mr, ms
        np.npy_intp i, j, k, l, s0, s1, s2, ds4, ds3, ds2
        np.npy_intp index, pindex, nvox, r, s, rr, ss
        np.npy_intp ss0, ss1, ss2 # strides
        np.npy_intp ss0d, ss1d, ss2d # strides
        np.npy_intp v0, v1, v2, v3 # vertices for mask
        np.npy_intp w0, w1, w2, w3 # vertices for data
        double l0, l1, l2, l3
        double res

    coords = coords.astype(np.float)
    mask = check_cast_bin8(mask)

    l0 = 0; l1 = 0; l2 = 0; l3 = 0

    pmask_shape = np.array(mask.shape) + 1
    s0, s1, s2 = pmask_shape[:3]
    pmask = np.zeros(pmask_shape, np.uint8)
    pmask[:-1, :-1, :-1] = mask

    fpmask = pmask.reshape(-1)
    fmask = mask.reshape(-1).astype(np.uint8)
    fcoords = coords.reshape((coords.shape[0], -1))

    # First do the interior contributions.
    # We first figure out which vertices, edges, triangles, tetrahedra
    # are uniquely associated with an interior voxel

    # The mask is copied into a larger array, hence it will have different
    # strides than the data
    cdef:
        np.ndarray[np.intp_t, ndim=1] strides
        np.ndarray[np.intp_t, ndim=1] dstrides
    strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
    dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
    ss0, ss1, ss2 = strides[0], strides[1], strides[2]
    ss0d, ss1d, ss2d = dstrides[0], dstrides[1], dstrides[2]
    verts = []
    for i in range(2):
        for j in range(2):
            for k in range(2):
                verts.append(ss0d * i + ss1d * j + ss2d * k)
    cvertices = np.array(sorted(verts), np.intp)

    union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
                             cube_with_strides_center((0,1,0), strides),
                             cube_with_strides_center((0,1,1), strides),
                             cube_with_strides_center((1,0,0), strides),
                             cube_with_strides_center((1,0,1), strides),
                             cube_with_strides_center((1,1,0), strides),
                             cube_with_strides_center((1,1,1), strides)])
    c = cube_with_strides_center((0,0,0), strides)
    m4 = np.array(list(c[4].difference(union[4])))
    m3 = np.array(list(c[3].difference(union[3])))
    m2 = np.array(list(c[2].difference(union[2])))

    d4 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m4[i]] for i in range(m4.shape[0])])
    d4 = np.hstack([m4, d4])
    ds4 = d4.shape[0]

    d3 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m3[i]] for i in range(m3.shape[0])])
    d3 = np.hstack([m3, d3])
    ds3 = d3.shape[0]

    d2 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m2[i]] for i in range(m2.shape[0])])
    d2 = np.hstack([m2, d2])
    ds2 = d2.shape[0]

    nvox = mask.size

    D = np.zeros((8,8))

    for i in range(s0-1):
        for j in range(s1-1):
            for k in range(s2-1):

                pindex = i*ss0+j*ss1+k*ss2
                index = i*ss0d+j*ss1d+k*ss2d
                for r in range(8):
                    rr = (index+cvertices[r]) % nvox
                    mr = fmask[rr]
                    for s in range(r+1):
                        res = 0
                        ss = (index+cvertices[s]) % nvox
                        ms = fmask[ss]
                        if mr * ms:
                            for l in range(fcoords.shape[0]):
                                res += fcoords[l,ss] * fcoords[l,rr]
                            D[r,s] = res
                            D[s,r] = res
                        else:
                            D[r,s] = 0
                            D[s,r] = 0

                for l in range(ds4):
                    v0 = pindex + d4[l,0]
                    w0 = d4[l,4]
                    m = fpmask[v0]
                    if m:
                        v1 = pindex + d4[l,1]
                        v2 = pindex + d4[l,2]
                        v3 = pindex + d4[l,3]
                        w1 = d4[l,5]
                        w2 = d4[l,6]
                        w3 = d4[l,7]

                        m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]

                        l3 = l3 + m * mu3_tet(D[w0,w0], D[w0,w1], D[w0,w2],
                                              D[w0,w3], D[w1,w1], D[w1,w2],
                                              D[w1,w3], D[w2,w2], D[w2,w3],
                                              D[w3,w3])

                        l2 = l2 - m * mu2_tet(D[w0,w0], D[w0,w1], D[w0,w2],
                                              D[w0,w3], D[w1,w1], D[w1,w2],
                                              D[w1,w3], D[w2,w2], D[w2,w3],
                                              D[w3,w3])

                        l1 = l1 + m * mu1_tet(D[w0,w0], D[w0,w1], D[w0,w2],
                                              D[w0,w3], D[w1,w1], D[w1,w2],
                                              D[w1,w3], D[w2,w2], D[w2,w3],
                                              D[w3,w3])

                        l0 = l0 - m

                for l in range(ds3):
                    v0 = pindex + d3[l,0]
                    w0 = d3[l,3]
                    m = fpmask[v0]
                    if m:
                        v1 = pindex + d3[l,1]
                        v2 = pindex + d3[l,2]
                        w1 = d3[l,4]
                        w2 = d3[l,5]

                        m = m * fpmask[v1] * fpmask[v2] 
                        l2 = l2 + m * mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
                                              D[w1,w1], D[w1,w2], D[w2,w2])

                        l1 = l1 - m * mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
                                              D[w1,w1], D[w1,w2], D[w2,w2])

                        l0 = l0 + m

                for l in range(ds2):
                    v0 = pindex + d2[l,0]
                    w0 = d2[l,2]
                    m = fpmask[v0]
                    if m:
                        v1 = pindex + d2[l,1]
                        w1 = d2[l,3]
                        m = m * fpmask[v1]
                        l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])

                        l0 = l0 - m

    # fpmask has the same sum as mask, but with predictable dtype
    l0 += fpmask.sum()
    return np.array([l0, l1, l2, l3])


def _convert_stride3(v, stride1, stride2):
    """
    Take a voxel, expressed as in index in stride1 and
    re-express it as an index in stride2
    """
    v0 = v // stride1[0]
    v -= v0 * stride1[0]
    v1 = v // stride1[1]
    v2 = v - v1 * stride1[1]
    return v0*stride2[0] + v1*stride2[1] + v2*stride2[2]


def _convert_stride2(v, stride1, stride2):
    """
    Take a voxel, expressed as in index in stride1 and
    re-express it as an index in stride2
    """
    v0 = v // stride1[0]
    v1 = v - v0 * stride1[0]
    return v0*stride2[0] + v1*stride2[1]


def _convert_stride1(v, stride1, stride2):
    """
    Take a voxel, expressed as in index in stride1 and
    re-express it as an index in stride2
    """
    v0 = v // stride1[0]
    return v0 * stride2[0]


def Lips2d(coords, mask):
    """ Estimate intrinsic volumes for 2d region in `mask` given `coords`

    Given a 2d `mask` and coordinates `coords`, estimate the intrinsic volumes
    of the masked region. The region is broken up into triangles / edges /
    vertices, which are included based on whether all voxels in the triangle /
    edge / vertex are in the mask or not.

    Parameters
    ----------
    coords : ndarray shape (N, i, j)
         Coordinates for the voxels in the mask. ``N`` will often be 2 (for 2
         dimensional coordinates), but can be any integer > 0
    mask : ndarray shape (i, j)
         Binary mask determining whether or not a voxel is in the mask.

    Returns
    -------
    mu : ndarray
        Array of intrinsic volumes [mu0, mu1, mu2], being, respectively:
        #. Euler characteristic
        #. 2 * mean caliper diameter
        #. Area.

    Notes
    -----
    We check whether `mask` is binary.

    Raises
    ------
    ValueError
        If any value in the mask is outside {0, 1}

    References
    ----------
    Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
      with an application to brain mapping."
      Journal of the American Statistical Association, 102(479):913-928.
    """
    if mask.shape != coords.shape[1:]:
        raise ValueError('shape of mask does not match coordinates')
    # if the data can be squeezed, we must use the lower dimensional function
    if mask.ndim == 1:
        value = np.zeros(3)
        coords = coords.reshape((coords.shape[0],) + mask.shape)
        value[:2] = Lips1d(coords, mask)
        return value

    cdef:
        # c-level versions of the arrays
        # 'flattened' coords (2d array)
        np.ndarray[np.float_t, ndim=2] fcoords
        np.ndarray[np.float_t, ndim=2] D
        # 'flattened' mask (1d array)
        np.ndarray[np.uint8_t, ndim=1] fmask
        np.ndarray[np.uint8_t, ndim=1] fpmask
        # d2 and d3 are lists of triangles associated to particular voxels in
        # the square
        np.ndarray[np.intp_t, ndim=2] d3
        np.ndarray[np.intp_t, ndim=2] d2
        np.ndarray[np.intp_t, ndim=1] cvertices
        # scalars
        np.npy_uint8 m, mr, ms
        np.npy_intp i, j, k, l, r, s, rr, ss, s0, s1
        np.npy_intp ds2, ds3, index, npix, pindex
        np.npy_intp ss0, ss1, ss0d, ss1d # strides
        np.npy_intp v0, v1, v2 # vertices
        np.npy_intp w0, w1, w2
        double l0, l1, l2
        double res

    coords = coords.astype(np.float)
    mask = check_cast_bin8(mask)

    l0 = 0; l1 = 0; l2 = 0

    pmask_shape = np.array(mask.shape) + 1
    pmask = np.zeros(pmask_shape, np.uint8)
    pmask[:-1, :-1] = mask

    s0, s1 = pmask.shape[:2]

    fpmask = pmask.reshape(-1)
    fmask = mask.reshape(-1).astype(np.uint8)
    fcoords = coords.reshape((coords.shape[0], -1))

    # First do the interior contributions.
    # We first figure out which vertices, edges, triangles, tetrahedra
    # are uniquely associated with an interior voxel

    # The mask is copied into a larger array, hence it will have different
    # strides than the data
    cdef:
        np.ndarray[np.intp_t, ndim=1] strides
        np.ndarray[np.intp_t, ndim=1] dstrides
    strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
    dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
    ss0, ss1 = strides[0], strides[1]
    ss0d, ss1d = dstrides[0], dstrides[1]
    verts = []
    for i in range(2):
        for j in range(2):
            verts.append(ss0d * i + ss1d * j)
    cvertices = np.array(sorted(verts), np.intp)

    union = join_complexes(*[cube_with_strides_center((0,1), strides),
                             cube_with_strides_center((1,0), strides),
                             cube_with_strides_center((1,1), strides)])

    c = cube_with_strides_center((0,0), strides)
    m3 = np.array(list(c[3].difference(union[3])))
    m2 = np.array(list(c[2].difference(union[2])))

    d3 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m3[i]] for i in range(m3.shape[0])])
    d3 = np.hstack([m3, d3])
    ds3 = d3.shape[0]

    d2 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m2[i]] for i in range(m2.shape[0])])
    d2 = np.hstack([m2, d2])
    ds2 = d2.shape[0]

    D = np.zeros((4,4))

    npix = mask.size

    for i in range(s0-1):
        for j in range(s1-1):
            pindex = i*ss0+j*ss1
            index = i*ss0d+j*ss1d
            for r in range(4):
                rr = (index+cvertices[r]) % npix
                mr = fmask[rr]
                for s in range(r+1):
                    res = 0
                    ss = (index+cvertices[s]) % npix
                    ms = fmask[ss]
                    if mr * ms:
                        for l in range(fcoords.shape[0]):
                            res += fcoords[l,ss] * fcoords[l,rr]
                            D[r, s] = res
                            D[s, r] = res
                    else:
                        D[r, s] = 0
                        D[s, r] = 0

            for l in range(ds3):
                v0 = pindex + d3[l,0]
                w0 = d3[l,3]
                m = fpmask[v0]
                if m:
                    v1 = pindex + d3[l,1]
                    v2 = pindex + d3[l,2]
                    w1 = d3[l,4]
                    w2 = d3[l,5]
                    m = m * fpmask[v1] * fpmask[v2]
                    l2 = l2 + mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
                                        D[w1,w1], D[w1,w2], D[w2,w2]) * m
                    l1 = l1 - mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
                                        D[w1,w1], D[w1,w2], D[w2,w2]) * m
                    l0 = l0 + m

            for l in range(ds2):
                v0 = pindex + d2[l,0]
                w0 = d2[l,2]
                m = fpmask[v0]
                if m:
                    v1 = pindex + d2[l,1]
                    w1 = d2[l,3]
                    m = m * fpmask[v1]
                    l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
                    l0 = l0 - m

    # fpmask has the same sum as mask, but with predictable dtype
    l0 += fpmask.sum()
    return np.array([l0,l1,l2])


def EC2d(mask):
    """ Compute Euler characteristic of 2D region in `mask`

    Given a 2d `mask`, compute the 0th intrinsic volume (Euler characteristic)
    of the masked region. The region is broken up into triangles / edges /
    vertices, which are included based on whether all voxels in the triangle /
    edge / vertex are in the mask or not.

    Parameters
    ----------
    mask : ndarray shape (i, j)
         Binary mask determining whether or not a voxel is in the mask.

    Returns
    -------
    mu0 : int
        Euler characteristic

    Notes
    -----
    We check whether `mask` is binary.

    Raises
    ------
    ValueError
        If any value in the mask is outside {0, 1}

    References
    ----------
    Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
      with an application to brain mapping."
      Journal of the American Statistical Association, 102(479):913-928.
    """
    cdef:
        # c-level versions of the array
        # 'flattened' mask (1d array)
        np.ndarray[np.uint8_t, ndim=1] fpmask
        # d2 and d3 are lists of triangles and tetrahedra
        # associated to particular voxels in the cuve
        np.ndarray[np.intp_t, ndim=2] d2
        np.ndarray[np.intp_t, ndim=2] d3
        # scalars
        np.uint8_t m
        np.npy_intp i, j, k, l, s0, s1, ds2, ds3, index
        np.ndarray[np.intp_t, ndim=1] strides
        np.npy_intp ss0, ss1 # strides
        np.npy_intp v0, v1 # vertices
        long l0 = 0

    mask = check_cast_bin8(mask)

    pmask_shape = np.array(mask.shape) + 1
    pmask = np.zeros(pmask_shape, np.uint8)
    pmask[:-1, :-1] = mask

    s0, s1 = pmask.shape[:2]

    fpmask = pmask.reshape(-1)

    strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
    ss0, ss1 = strides[0], strides[1]

    # First do the interior contributions.
    # We first figure out which vertices, edges, triangles, tetrahedra
    # are uniquely associated with an interior voxel
    union = join_complexes(*[cube_with_strides_center((0,1), strides),
                             cube_with_strides_center((1,0), strides),
                             cube_with_strides_center((1,1), strides)])
    c = cube_with_strides_center((0,0), strides)

    d3 = np.array(list(c[3].difference(union[3])))
    d2 = np.array(list(c[2].difference(union[2])))

    ds2 = d2.shape[0]
    ds3 = d3.shape[0]

    for i in range(s0-1):
        for j in range(s1-1):
            index = i*ss0+j*ss1
            for l in range(ds3):
                v0 = index + d3[l,0]
                m = fpmask[v0]
                if m and v0:
                    v1 = index + d3[l,1]
                    v2 = index + d3[l,2]
                    m = m * fpmask[v1] * fpmask[v2]
                    l0 = l0 + m

            for l in range(ds2):
                v0 = index + d2[l,0]
                m = fpmask[v0]
                if m:
                    v1 = index + d2[l,1]
                    m = m * fpmask[v1]
                    l0 = l0 - m

    # fpmask has the same sum as mask, but with predictable dtype
    l0 += fpmask.sum()
    return l0


def Lips1d(coords, mask):
    """ Estimate intrinsic volumes for 1D region in `mask` given `coords`

    Given a 1d `mask` and coordinates `coords`, estimate the intrinsic volumes
    of the masked region. The region is broken up into edges / vertices, which
    are included based on whether all voxels in the edge / vertex are in the
    mask or not.

    Parameters
    ----------
    coords : ndarray shape (N, i)
         Coordinates for the voxels in the mask. ``N`` will often be 1 (for 1
         dimensional coordinates), but can be any integer > 0
    mask : ndarray shape (i,)
         Binary mask determining whether or not a voxel is in the mask.

    Returns
    -------
    mu : ndarray
        Array of intrinsic volumes [mu0, mu1], being, respectively:
        #. Euler characteristic
        #. Line segment length

    Notes
    -----
    We check whether `mask` is binary.

    Raises
    ------
    ValueError
        If any value in the mask is outside {0, 1}

    References
    ----------
    Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
      with an application to brain mapping."
      Journal of the American Statistical Association, 102(479):913-928.
    """
    if mask.shape[0] != coords.shape[1]:
        raise ValueError('shape of mask does not match coordinates')
    cdef:
        np.ndarray[np.uint8_t, ndim=1] mask_c
        np.ndarray[np.float_t, ndim=2] coords_c
        np.uint8_t m, mr, ms
        np.npy_intp i, l, r, s, rr, ss, s0, index
        double l0, l1
        double res

    coords_c = coords.astype(np.float)
    mask_c = check_cast_bin8(mask)

    l0 = 0; l1 = 0
    s0 = mask_c.shape[0]
    D = np.zeros((2,2))

    for i in range(s0):
        for r in range(2):
            rr = (i+r) % s0
            mr = mask_c[rr]
            for s in range(r+1):
                res = 0
                ss = (i+s) % s0
                ms = mask_c[ss]
                if mr * ms * ((i+r) < s0) * ((i+s) < s0):
                    for l in range(coords_c.shape[0]):
                        res += coords_c[l,ss] * coords_c[l,rr]
                        D[r,s] = res
                        D[s,r] = res
                else:
                    D[r,s] = 0
                    D[s,r] = 0

        m = mask_c[i]
        if m:
            m = m * (mask_c[(i+1) % s0] * ((i+1) < s0))
            l1 = l1 + m * mu1_edge(D[0,0], D[0,1], D[1,1])
            l0 = l0 - m

    # mask_c has the same sum as mask, but with predictable dtype
    l0 += mask_c.sum()
    return np.array([l0, l1])


def EC1d(mask):
    """ Compute Euler characteristic for 1d `mask`

    Given a 1d mask `mask`, compute the 0th intrinsic volume (Euler
    characteristic) of the masked region. The region is broken up into edges /
    vertices, which are included based on whether all voxels in the edge /
    vertex are in the mask or not.

    Parameters
    ----------
    mask : ndarray shape (i,)
         Binary mask determining whether or not a voxel is in the mask.

    Returns
    -------
    mu0 : int
        Euler characteristic

    Notes
    -----
    We check whether the array mask is binary.

    The 3d cubes are triangulated into 6 tetrahedra of equal volume, as
    described in the reference below.

    Raises
    ------
    ValueError
        If any value in the mask is outside {0, 1}

    References
    ----------
    Taylor, J.E. & Worsley, K.J. (2007). "Detecting sparse signal in random fields,
      with an application to brain mapping."
      Journal of the American Statistical Association, 102(479):913-928.
    """
    cdef:
        np.ndarray[np.uint8_t, ndim=1] mask_c
        np.uint8_t m
        np.npy_intp i, s0
        double l0 = 0

    mask_c = check_cast_bin8(mask)
    s0 = mask_c.shape[0]
    for i in range(s0):
        m = mask_c[i]
        if m:
            m = m * (mask_c[(i+1) % s0] * ((i+1) < s0))
            l0 = l0 - m

    # mask_c has the same sum as mask, but with predictable dtype
    l0 += mask_c.sum()
    return l0



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Image interpolators using ndimage.
"""
from __future__ import absolute_import

import tempfile

from distutils.version import LooseVersion as LV

import numpy as np

import scipy
from scipy import ndimage

from ..fixes.scipy.ndimage import map_coordinates
from ..utils import seq_prod


# Earlier versions of Scipy don't have mode for spline_filter
SCIPY_VERSION = LV(scipy.__version__)
SPLINE_FILTER_HAS_MODE = SCIPY_VERSION >= LV('1.2')
# Fixes in interpolation in scipy >= 1.6 force pre-padding
# in knot calculations.
SPLINE_FILTER_NEEDS_PAD = SCIPY_VERSION >= LV('1.6')


class ImageInterpolator(object):
    """ Interpolate Image instance at arbitrary points in world space

    The resampling is done with ``scipy.ndimage``.
    """

    # Padding for prefilter calculation in 'nearest' and 'grid-constant' mode.
    # See: https://github.com/scipy/scipy/issues/13600
    n_prepad_if_needed = 12 if SPLINE_FILTER_NEEDS_PAD else 0

    def __init__(self, image, order=3, mode='constant', cval=0.0):
        """
        Parameters
        ----------
        image : Image
           Image to be interpolated.
        order : int, optional
           order of spline interpolation as used in ``scipy.ndimage``.
           Default is 3.
        mode : str, optional
           Points outside the boundaries of the input are filled according to
           the given mode ('constant', 'nearest', 'reflect' or 'wrap'). Default
           is 'constant'.
        cval : scalar, optional
           Value used for points outside the boundaries of the input if
           mode='constant'. Default is 0.0.
        """
        # order and mode are read-only to allow pre-calculation of spline
        # filters.
        self.image = image
        self._order = order
        self._mode = mode
        self.cval = cval
        self._datafile = None
        self._n_prepad = 0  # Non-zero for 'nearest' and 'grid-constant'
        self._buildknots()

    @property
    def mode(self):
        """ Mode is read-only
        """
        return self._mode

    @property
    def order(self):
        """ Order is read-only
        """
        return self._order

    def _buildknots(self):
        data = np.nan_to_num(self.image.get_data()).astype(np.float64)
        if self.order > 1:
            if self.mode in ('nearest', 'grid-constant'):
                # See: https://github.com/scipy/scipy/issues/13600
                self._n_prepad = self.n_prepad_if_needed
                if self._n_prepad != 0:
                    data = np.pad(data, self._n_prepad, mode='edge')
            kwargs = dict(order=self.order)
            if SPLINE_FILTER_HAS_MODE:
                kwargs['mode'] = self.mode
            data = ndimage.spline_filter(data, **kwargs)
        self._datafile = tempfile.TemporaryFile()
        data.tofile(self._datafile)
        self._data = np.memmap(self._datafile,
                               dtype=data.dtype,
                               mode='r+',
                               shape=data.shape)
        del(data)

    def evaluate(self, points):
        """ Resample image at points in world space

        Parameters
        ----------
        points : array
           values in self.image.coordmap.output_coords.  Each row is a point.

        Returns
        -------
        V : ndarray
           interpolator of self.image evaluated at points
        """
        points = np.array(points, np.float64)
        output_shape = points.shape[1:]
        points.shape = (points.shape[0], seq_prod(output_shape))
        cmapi = self.image.coordmap.inverse()
        voxels = cmapi(points.T).T + self._n_prepad
        V = map_coordinates(self._data,
                            voxels,
                            order=self.order,
                            mode=self.mode,
                            cval=self.cval,
                            prefilter=self.order < 2)
        # ndimage.map_coordinates returns a flat array,
        # it needs to be reshaped to the original shape
        V.shape = output_shape
        return V



#!/usr/bin/env python3
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
This script generates a noisy multi-subject activation image dataset
and applies the Bayesian structural analysis on it

Requires matplotlib

Author : Bertrand Thirion, 2009-2013
"""
print(__doc__)

import numpy as np
import scipy.stats as st

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

import nipy.labs.utils.simul_multisubject_fmri_dataset as simul
from nipy.labs.spatial_models.bayesian_structural_analysis import\
    compute_landmarks
from nipy.labs.spatial_models.discrete_domain import grid_domain_from_shape

    
def display_landmarks_2d(landmarks, hrois, stats):
    """ Plots the landmarks and associated rois as images"""
    shape = stats[0].shape
    n_subjects = len(stats)
    lmax = 0
    grp_map, density = np.zeros(shape), np.zeros(shape)
    if landmarks is not None:
        domain = landmarks.domain
        grp_map = landmarks.map_label(domain.coord, .8, sigma).reshape(shape)
        density = landmarks.kernel_density(k=None, coord=domain.coord,
                                           sigma=sigma).reshape(shape)
        lmax = landmarks.k + 2
        
    # Figure 1: input data    
    fig_input = plt.figure(figsize=(8, 3.5))
    fig_input.text(.5,.9, "Input activation maps", ha='center')
    vmin, vmax = stats.min(), stats.max()
    for subject in range(n_subjects):
        plt.subplot(n_subjects / 5, 5, subject + 1)
        plt.imshow(stats[subject], interpolation='nearest',
                   vmin=vmin, vmax=vmax)
        plt.axis('off')

    # Figure 2: individual hrois
    fig_output = plt.figure(figsize=(8, 3.5))
    fig_output.text(.5, .9, "Individual landmark regions", ha="center")
    for subject in range(n_subjects):
        plt.subplot(n_subjects / 5, 5, subject + 1)
        lw = - np.ones(shape)
        if hrois[subject].k > 0:
            nls = hrois[subject].get_roi_feature('label')
            nls[nls == - 1] = np.size(landmarks) + 2
            for k in range(hrois[subject].k):
                np.ravel(lw)[hrois[subject].label == k] = nls[k]

        plt.imshow(lw, interpolation='nearest', vmin=-1, vmax=lmax)
        plt.axis('off')
        
    # Figure 3: Group-level results
    plt.figure(figsize=(6, 3))

    plt.subplot(1, 2, 1)
    plt.imshow(grp_map, interpolation='nearest', vmin=-1, vmax=lmax)
    plt.title('group-level position 80% \n confidence regions', fontsize=10)
    plt.axis('off')
    plt.colorbar(shrink=.8)

    plt.subplot(1, 2, 2)
    plt.imshow(density, interpolation='nearest')
    plt.title('Spatial density under h1', fontsize=10)
    plt.axis('off')
    plt.colorbar(shrink=.8)


###############################################################################
# Main script
###############################################################################

# generate the data
n_subjects = 10
shape = (60, 60)
pos = np.array([[12, 14],
                [20, 20],
                [30, 20]])
ampli = np.array([5, 7, 6])
sjitter = 1.0
stats = simul.surrogate_2d_dataset(n_subj=n_subjects, shape=shape, pos=pos,
                                   ampli=ampli, width=5.0)

# set various parameters
threshold = float(st.t.isf(0.01, 100))
sigma = 4. / 1.5
prevalence_threshold = n_subjects * .25
prevalence_pval = 0.9
smin = 5
algorithm = 'co-occurrence' #  'density'

domain = grid_domain_from_shape(shape) 

# get the functional information
stats_ = np.array([np.ravel(stats[k]) for k in range(n_subjects)]).T
    
# run the algo
landmarks, hrois = compute_landmarks(
    domain, stats_, sigma, prevalence_pval, prevalence_threshold, 
    threshold, smin, method='prior', algorithm=algorithm)

display_landmarks_2d(landmarks, hrois, stats)
if landmarks is not None:
    landmarks.show()

plt.show()




#!/usr/bin/env python3
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
from os.path import join as pjoin, exists
from glob import glob
from distutils import log

# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if exists('MANIFEST'): os.remove('MANIFEST')

# Always use setuptools
import setuptools

from setup_helpers import (generate_a_pyrex_source, get_comrec_build,
                           cmdclass, INFO_VARS)

# monkey-patch numpy distutils to use Cython instead of Pyrex
from numpy.distutils.command.build_src import build_src
build_src.generate_a_pyrex_source = generate_a_pyrex_source

# Add custom commit-recording build command
from numpy.distutils.command.build_py import build_py as _build_py
cmdclass['build_py'] = get_comrec_build('nipy', _build_py)

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)
    # The quiet=True option will silence all of the name setting warnings:
    # Ignoring attempt to set 'name' (from 'nipy.core' to 
    #    'nipy.core.image')
    # Robert Kern recommends setting quiet=True on the numpy list, stating
    # these messages are probably only used in debugging numpy distutils.
    config.get_version(pjoin('nipy', 'info.py')) # sets config.version
    config.add_subpackage('nipy', 'nipy')
    return config

################################################################################

# Hard and soft dependency checking
DEPS = (
    ('numpy', INFO_VARS['NUMPY_MIN_VERSION'], 'setup_requires'),
    ('scipy', INFO_VARS['SCIPY_MIN_VERSION'], 'install_requires'),
    ('nibabel', INFO_VARS['NIBABEL_MIN_VERSION'], 'install_requires'),
    ('sympy', INFO_VARS['SYMPY_MIN_VERSION'], 'install_requires'),
)

requirement_kwargs = {'setup_requires': [], 'install_requires': []}
for name, min_ver, req_type in DEPS:
    new_req = '{0}>={1}'.format(name, min_ver)
    requirement_kwargs[req_type].append(new_req)


################################################################################
# commands for installing the data
from numpy.distutils.command.install_data import install_data
from numpy.distutils.command.build_ext import build_ext

def data_install_msgs():
    # Check whether we have data packages
    try:  # Allow setup.py to run without nibabel
        from nibabel.data import datasource_or_bomber
    except ImportError:
        log.warn('Cannot check for optional data packages: see: '
                 'http://nipy.org/nipy/users/install_data.html')
        return
    DATA_PKGS = INFO_VARS['DATA_PKGS']
    templates = datasource_or_bomber(DATA_PKGS['nipy-templates'])
    example_data = datasource_or_bomber(DATA_PKGS['nipy-data'])
    for dpkg in (templates, example_data):
        if hasattr(dpkg, 'msg'): # a bomber object, warn
            log.warn('%s\n%s' % ('_'*80, dpkg.msg))


class MyInstallData(install_data):
    """ Subclass the install_data to generate data install warnings if necessary
    """
    def run(self):
        install_data.run(self)
        data_install_msgs()


class MyBuildExt(build_ext):
    """ Subclass the build_ext to generate data install warnings if
        necessary: warn at build == warn early
        This is also important to get a warning when run a 'develop'.
    """
    def run(self):
        build_ext.run(self)
        data_install_msgs()


cmdclass['install_data'] = MyInstallData
cmdclass['build_ext'] = MyBuildExt

################################################################################


def main():
    from numpy.distutils.core import setup

    setup(name=INFO_VARS['NAME'],
          maintainer=INFO_VARS['MAINTAINER'],
          maintainer_email=INFO_VARS['MAINTAINER_EMAIL'],
          description=INFO_VARS['DESCRIPTION'],
          long_description=INFO_VARS['LONG_DESCRIPTION'],
          url=INFO_VARS['URL'],
          download_url=INFO_VARS['DOWNLOAD_URL'],
          license=INFO_VARS['LICENSE'],
          classifiers=INFO_VARS['CLASSIFIERS'],
          author=INFO_VARS['AUTHOR'],
          author_email=INFO_VARS['AUTHOR_EMAIL'],
          platforms=INFO_VARS['PLATFORMS'],
          # version set by config.get_version() above
          configuration = configuration,
          cmdclass = cmdclass,
          tests_require=['nose>=1.0'],
          test_suite='nose.collector',
          zip_safe=False,
          entry_points={
              'console_scripts': [
                  'nipy_3dto4d = nipy.cli.img3dto4d:main',
                  'nipy_4dto3d = nipy.cli.img4dto3d:main',
                  'nipy_4d_realign = nipy.cli.realign4d:main',
                  'nipy_tsdiffana = nipy.cli.tsdiffana:main',
                  'nipy_diagnose = nipy.cli.diagnose:main',
              ],
          },
          **requirement_kwargs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Example running a parcel-based second-level analysis from a set of
first-level effect images.

This script takes as input a directory path that contains first-level
images in nifti format, as well as a group mask image and a
parcellation image (such as the AAL atlas, 'ROI_MNI_V4.nii', see
http://www.gin.cnrs.fr/spip.php?article217). All images are assumed to
be in a common reference space, e.g. the MNI/Talairach space.

It outputs three images:

* tmap.nii.gz, a `t-statistic` image similar to a SPM-like second-level
  t-map, except it is derived under an assumption of localization
  uncertainty in reference space.

* parcel_mu.nii.gz, an image that maps each voxel to the estimated
  population effect in the parcel it belongs to.

* parcel_prob.nii.gz, an image that maps each voxel to the probability
  that the population effect in the parcel it belongs to is
  positive-valued.

See the `nipy.algorithms.group.ParcelAnalysis` class for more general
usage information.
"""
from os.path import join
from glob import glob
from nipy import load_image
from nipy.algorithms.group import parcel_analysis
from nipy.externals.argparse import ArgumentParser

# Parse command line
description = 'Run a parcel-based second-level analysis from a set of\
first-level effect images.'

parser = ArgumentParser(description=description)
parser.add_argument('con_path', metavar='con_path',
                    help='directory where 1st-level images are to be found')
parser.add_argument('msk_file', metavar='msk_file',
                    help='group mask file')
parser.add_argument('parcel_file', metavar='parcel_file',
                    help='parcellation image file')
args = parser.parse_args()

# Load first-level images
con_files = glob(join(args.con_path, '*.nii'))
con_imgs = [load_image(f) for f in con_files]

# Load group mask
msk_img = load_image(args.msk_file)

# Load parcellation
parcel_img = load_image(args.parcel_file)

# Run parcel analysis and write output images in the current directory
effect_img, proba_img = parcel_analysis(con_imgs, parcel_img,
                                        msk_img=msk_img, fwhm=8,
                                        res_path='.')
